// Part of SimCoupe - A SAM Coup emulator
//
// Parallel.cpp: Win32 parallel interface
//
//  Copyright (c) 1999-2001  Simon Owen
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

// Notes:
//  The main part of the DAC implementation is in Sound.cpp

#include "SimCoupe.h"
#include <winspool.h>   // for print spooler definitions
#include "Parallel.h"

#include "IO.h"
#include "Options.h"
#include "Sound.h"
#include "Util.h"

////////////////////////////////////////////////////////////////////////////////


CPrinterDevice::CPrinterDevice ()
    : m_hPrinter(INVALID_HANDLE_VALUE), m_nPrinter(0), m_bPrint(0), m_bStatus(0), m_fFailed(false)
{
}

CPrinterDevice::~CPrinterDevice ()
{
    // Make sure any existing job has finished
    Close();
}


BYTE CPrinterDevice::In (WORD wPort_)
{
    return (wPort_ & 1) ? m_bStatus : 0;
}

void CPrinterDevice::Out (WORD wPort_, BYTE bVal_)
{
    // If the write is to the data port, save the byte for later
    if (!(wPort_ & 1))
        m_bPrint = bVal_;

    else
    {
        // If the printer is being strobed, write the data byte
        if ((m_bStatus & 0x01) && !(bVal_ & 0x01))
            Write(m_bPrint);

        // Update the status with the strobe bit
        (m_bStatus &= ~0x01) |= (bVal_ & 0x01);
    }
}


bool CPrinterDevice::Open ()
{
    Close();

    PRINTER_DEFAULTS pd = { "RAW", NULL, PRINTER_ACCESS_USE };

    if (OpenPrinter(const_cast<char*>(GetOption(printerdev)), &m_hPrinter, &pd))
    {
        DOC_INFO_1 docinfo;
        docinfo.pDocName = "SimCoupe print";
        docinfo.pOutputFile = NULL;
        docinfo.pDatatype = "RAW";

        // Start the job
        if (StartDocPrinter(m_hPrinter, 1, reinterpret_cast<BYTE*>(&docinfo)) && StartPagePrinter(m_hPrinter))
        {
            // Signal the printer is ready to receive data
            m_bStatus |= 0x81;
            return true;
        }
    }

    // Only complain if we've not already done so, to prevent it happening for every byte printed!
    if (!m_fFailed)
    {
        Message(msgWarning, "Failed to start print job!");
        m_fFailed = true;
    }

    // Clean up anything that was opened
    Close();
    return false;
}

void CPrinterDevice::Close ()
{
    if (IsOpen())
    {
        Flush();

        EndPagePrinter(m_hPrinter);
        EndDocPrinter(m_hPrinter);
        ClosePrinter(m_hPrinter);

        m_hPrinter = INVALID_HANDLE_VALUE;

        m_bStatus = 0;
    }
}

void CPrinterDevice::Write (BYTE bPrint_)
{
    // Start a new job if there isn't an existing one
    if (IsOpen() || Open())
    {
        if (m_nPrinter < sizeof m_abPrinter)
        {
            m_abPrinter[m_nPrinter++] = bPrint_;

            if (m_nPrinter >= sizeof m_abPrinter)
                Flush();
        }
    }
}

void CPrinterDevice::Flush ()
{
    if (m_nPrinter && IsOpen())
    {
        DWORD dwWritten;
        if (!WritePrinter(m_hPrinter, m_abPrinter, m_nPrinter, &dwWritten))
        {
            TRACE("!!! Error writing to printer (%#08lx)!\n");
            Close();
        }
        else if (m_nPrinter != dwWritten)
            TRACE("!!! Only printed %lu of %d bytes!\n", dwWritten, m_nPrinter);
    }

    m_nPrinter = 0;
}


////////////////////////////////////////////////////////////////////////////////


void CMonoDACDevice::Out (WORD wPort_, BYTE bVal_)
{
    // If the write is to the data port, send it to the DAC
    if (!(wPort_ & 1))
        Sound::OutputDAC(bVal_);
}

////////////////////////////////////////////////////////////////////////////////


void CStereoDACDevice::Out (WORD wPort_, BYTE bVal_)
{
    wPort_ &= 3;

    // Sample value?
    if (!(wPort_ & 1))
        m_bVal = bVal_;

    // Output to left or right channel
    else if (bVal_ & 1)
        Sound::OutputDACLeft(m_bVal);
    else
        Sound::OutputDACRight(m_bVal);
}
